import numpy as np
import numpy.linalg as alg
import file_reader
import operator
import sys
import args

def print_result(c, vals, input_data):
    print "Optimal solution was found:"
    for i in range(0, input_data.get_N()):
        print 'x%s = %s' % (i, round(vals[i], 6), )
    print "F =", (c * values[range(0, input_data.get_N()), 0])[0, 0]


def pretty_print(basis_solution, present_indices, B, u, ABj, delta, teta, delta_index, f_val):
    upper_line = ["", "" if f_val is None else str(f_val) ] + [str(u[0, val]) for val in range(0, u.shape[1])] + [('Delta(%d) = %s' % (delta_index, delta, ))]
    middle_lines = list()
    for i in range(0, B.shape[0]):
        middle_lines.append(['x`' + str(present_indices[i]), str(basis_solution[i, 0])] + [str(B[i, j]) for j in range(0, B.shape[1])] + [str(ABj[i, 0])])

    all_lines = [upper_line] + middle_lines    
    gaps = list()
    for i in range(0, len(upper_line)):
        gaps.append(max([len(all_lines[k][i]) for k in range(0, len(all_lines))]))

    
    total = sum(gaps)
    divider_length = total + 3 * len(gaps) + 1
    print '_' * divider_length +  '\n'
    for i in range(0, len(all_lines)):
        s = "|"
        for j in range(0, len(all_lines[i])):
            s = s + ((' {:' + str(gaps[j]) + 's} ').format(all_lines[i][j])) + '|'
        print s
        print '_' * divider_length +  '\n'

def check(input_data, basis_solution):
    equations = input_data.get_Equations()
    equations_matrix = list()
    print "Current solution: \n", basis_solution
    for i in range(0, len(equations)):
        eq = equations[i][:-1]
        for j in range(0, input_data.get_M()):
            if i == j:
                eq.append(-1)
            else:
                eq.append(0)
        equations_matrix.append(eq)
    mistakes = list()
    for i in range(0, len(equations)):
        total = 0
        s = ""
        for k in range(0, len(equations_matrix[i])):
            if not s == "":
                s = s + " + "
            x = equations_matrix[i][k]
            y = basis_solution[k]
            s = s + ('%s * %s' % (x if x>= 0 else ('(' + str(x) + ')') , y[0, 0] if y>= 0 else ('(' + str('%.3f' % round(y[0, 0], 3)) + ')'), ))
            total = total + equations_matrix[i][k] * y[0, 0]
        s = s + " = " + str('%.3f' % round(total, 3))
        s = s + " expected 0\n"
        if not round(total) == 0:
            mistakes.append((i, total))
        print s

    for i in range(0, len(mistakes)):
        print 'Mistake at ', mistakes[i][0], 'equation: found', mistakes[i][1], "expected 0"
    print "\n"

def replace_column(matrix, i, j):
    m = np.matrix(np.copy(matrix))
    m[:,i] = j
    return m

def estimate(input_data, target_function, B, current_basis):
    equations = np.matrix(input_data.get_Equations())
    estimation=list()
    temp = target_function * B
#    print target_function
    for i in range(0, input_data.get_N()):
#        if not current_basis[i, 0] == 0:
#            continue
        est = temp * extract_as_int(equations,i)
        estimation.append((i, int(est.item(0))))
    return estimation

def estimate2(input_data, target_function, B, c, basis_indices):
    equations = np.matrix(input_data.get_Equations())
    estimation=list()
    for i in range(0, input_data.get_N()):
        if i in basis_indices: continue
        est = target_function * extract_as_int(equations,i) - int(alg.det(B)) * c[0, i]
        estimation.append((i, int(round(est.item(0)))))
    return estimation

def filter_estimates(input_data, estimation, detA_b):
    available_for_improvement = list()
    for i in range(0, len(estimation)):
        border = input_data.get_bound(i)
        if border.isL() and estimation[i][1] < 0:
#            print "Skipped estimate for ", "x" + str(i), "from L and value", (estimation[i][0], int(estimation[i][1]))
            continue
        if border.isU() and estimation[i][1] > 0:
#            print "Skipped estimate for ", "x" + str(i), "from U and value", (estimation[i][0], int(estimation[i][1]))
            continue
        if border.isZ() and estimation[i][1] == 0:
#            print "Skipped estimate for ", "x" + str(i), "from Z and value", (estimation[i][0], int(estimation[i][1]))
            continue
        available_for_improvement.append((0, estimation[i][0], estimation[i][1]) )

    return available_for_improvement


def choose_estimation(available):
    return max(available_for_improvement,key=operator.itemgetter(2))

def extract_column(matrix, col):
    return matrix[:,col]

def matrix_to_1D_array(matrix):
    return np.squeeze(np.asarray(matrix))

def calculate_tetas(input_data, idx, basis_solution, current_working_indices, BAj, detA_b):
    teta_1 = list()
    teta_2 = list()
    for i in range(0, len(BAj)):
        cur = current_working_indices[i]
        border = input_data.get_bound(cur)
#        if cur < input_data.get_N():
#            continue
        x_i = basis_solution.item(i)
        l_i = border.get_lower_value()
        u_i = border.get_upper_value()
        z_i = BAj[i]
#        print i, x_i, l_i, u_i, z_i
        if BAj[i]> 0 and idx[2] > 0:
            t = (x_i - abs(detA_b) * l_i) / z_i
            teta_1.append((i, cur, t, idx[1], 1))
        elif BAj[i] < 0 and idx[2] > 0:
            t = (x_i - abs(detA_b) * u_i) / z_i
            teta_2.append((i, cur, t, idx[1], 2))
        elif BAj[i]> 0 and idx[2] < 0:
            t = (u_i - abs(detA_b) * x_i) / z_i
            teta_1.append((i, cur, t, idx[1], 1))
        elif BAj[i]< 0 and idx[2] < 0:
            t = (l_i - abs(detA_b) * x_i) / z_i
            teta_2.append((i, cur, t, idx[1], 2))

    return (teta_1, teta_2)

def choose_teta(teta_1, teta_2):
    return min(teta_1 + teta_2,key=operator.itemgetter(2))


def create_q_matrix(base):
    B = base[:,[i for i in range(0, base.shape[0])]]
    Q = np.zeros(base.shape, int)
    for i in range(0, base.shape[0]):
        for j in range(0, base.shape[1]):        
             Q[i,j] = int(round(alg.det(replace_column(B, i, base[:,j:j+1]))))
    return Q

def make_q_rotate(base, Q, replacement):
    B = base[:,[i for i in range(0, base.shape[0])]]
    detB = alg.det(B)
    Q_rotated = np.zeros(Q.shape, int)
    # ! we can calculate just middle
    for i in range(0, Q.shape[0]):
        for j in range(0, Q.shape[1]):
            if i == replacement:
                Q_rotated[i,j] = int(round(Q[i,j]))
            else:
                Q_rotated[i,j] = int(round((Q[i, j] * Q[replacement, Q.shape[1]-1] - Q[replacement, j] * Q[i, Q.shape[1]-1]) / detB))
    return Q_rotated

def extract_B_from_q(input_data, Q):
    return Q[:,[i for i in range(input_data.get_M(), 2 * input_data.get_M())]]
#    return Q[:,[i for i in range(0, input_data.get_M())]]

def calculate_d(input_data, Q0, teta_basis_index):
    AB_apostr = Q0[:,[i for i in range(input_data.get_M(), 2 * input_data.get_M())]]
    p = np.matrix(Q0[:,-1]).transpose()
    d_apostr = AB_apostr * p
    return d_apostr, d_apostr[teta_basis_index, 0]

def extract_as_int(e, i):
    fl = equations[:,i:i+1]
    return fl.astype(int)


fname = raw_input('Enter a filename: ')
#print fname
if fname is None or fname == '':
    fname = 'data.txt'

print 'Reading', fname, '...\n'
input_data = file_reader.read_input_data(fname)
for i in range(0, input_data.get_M()):
    input_data.add_bound(input_data.get_N() + i, (file_reader.MORE, float("-inf")), (file_reader.LESS_OR_EQ, input_data.get_Equations()[i][-1]))


a = np.concatenate((np.matrix([t[:-1] for t in input_data.get_Equations()]), np.identity(input_data.get_M()) * -1), axis=1)

A_b = np.diag([-1 for i in range(0, input_data.get_M())])
A_b_ = np.diag([1 for i in range(0, input_data.get_M())])
A_b_0 = A_b_ * -1

basis_indices = [i + input_data.get_N() for i in range(0, input_data.get_M())]
values = np.matrix([0.0 for i in range(0, input_data.get_N() + input_data.get_M())]).transpose()
x_b = np.matrix([0 for i in range(0, input_data.get_M())]).transpose()
basis = np.matrix([0 for i in range(0, input_data.get_M() + input_data.get_N())])
u = np.matrix([-1 for i in range(0, input_data.get_M())], int)
equations = np.matrix(input_data.get_Equations())

c = np.matrix(input_data.get_Target())

AbAj = np.zeros(input_data.get_M())
iteration = 1
direction_backward = False
while True:
    print "=== Phase 1", "Iteration", iteration, ".\n"
#    print A_b_0

    estimation = estimate(input_data, u, A_b_0, values)
    print "Delta:", [(t[0], t[1]) for t in estimation]
    available_for_improvement = filter_estimates(input_data, estimation, alg.det(A_b))

    if len(available_for_improvement) == 0:
        iteration = 1
        direction_backward = True
        break

    
#    print "Working estimates: ", [(t[1], int(t[2])) for t in available_for_improvement]
    max_estimation_index = choose_estimation(available_for_improvement)
    print "Chosen:", "x" + str(max_estimation_index[1]), " with delta", max_estimation_index[2]

#    AbAj_old = np.zeros(input_data.get_M()) if AbAj is None else AbAj


    teta_1, teta_2 = calculate_tetas(input_data, max_estimation_index, x_b, basis_indices, matrix_to_1D_array(A_b_0 * extract_column(equations, max_estimation_index[1])), alg.det(A_b))
    print "Theta:", [("x" + str(i[1]), round(i[2], 5)) for i in teta_1], "and", [("x" + str(i[1]), round(i[2], 5)) for i in teta_2]

    if len(teta_1 + teta_2) == 0:
        iteration = 1
        direction_backward = True
        break
    elif all([i[2] == float('inf') for i in (teta_1 + teta_2)]):
        print "No non-infinitive theta was found."
        sys.exit("No solution have been found: objective function is not limited.")

    teta = choose_teta(teta_1, teta_2)
    print "Chosen theta:", teta[2], "for", "x" + str(teta[1])
    AbAj = A_b_0 * extract_as_int(equations, max_estimation_index[1])

    x_to_remove = basis_indices[teta[0]]
    x_to_add = max_estimation_index[1]
    print "Replacing", "x" + str(x_to_remove), " with ", "x" + str(x_to_add)
    Q_base = np.concatenate((A_b, np.identity(A_b.shape[0], int), extract_as_int(equations, x_to_add)), axis=1	)
    Q0 = create_q_matrix(Q_base)
    Q1 = make_q_rotate(Q_base, Q0, teta[0])
    print "Building Q-matrix..."
    print Q_base
    print "..."
    print Q0
    print "Rotating column", teta[0], "..."
    print Q1
    print "Next basis matrix determinant =", Q1[teta[0],Q1.shape[1] - 1]

    pretty_print(x_b, basis_indices, A_b_0, u, AbAj, max_estimation_index[2], teta[2], max_estimation_index[1], None)
    
    A_b[:,teta[0]:teta[0]+1] = extract_as_int(equations, x_to_add)
    A_b_ = Q1[:,[input_data.get_M() + i for i in range(0, input_data.get_M())]]
    if teta[4] == 1 and max_estimation_index[2] < 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_upper_value()
    elif teta[4] == 1 and max_estimation_index[2] > 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_lower_value()
    elif teta[4] == 2 and max_estimation_index[2] < 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_lower_value()
    elif teta[4] == 2 and max_estimation_index[2] > 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_upper_value()

    basis_indices[teta[0]] = max_estimation_index[1]
    
    non_basis_values = list()
    for i in range(0, values.shape[0]):
        if i not in basis_indices:
            if i >= input_data.get_N():
                non_basis_values.append(values[i, 0])
            else:
                non_basis_values.append(values[i, 0])
        else:
            non_basis_values.append(0)
    non_basis_values = np.matrix(non_basis_values).transpose()
    x_b_ = - int(np.sign(alg.det(A_b))) * A_b_ * (a * non_basis_values).astype(int)
    for i in range(0, x_b_.shape[0]):
        values[basis_indices[i]] = x_b_[i, 0] / abs(alg.det(A_b))
    x_b = x_b_
    A_b_0 = int(np.sign(alg.det(A_b))) * A_b_
    u[0, teta[0]] = 0
#    check(input_data, values)
    iteration = iteration + 1


for var in basis_indices:
    if var >= input_data.get_N():
        print "Synthetic variable is present:", "x" + str(var)
        if AbAj[basis_indices.index(var), 0] == 0:
            sys.exit("Linear system is inconsistent.")
        else:
            sys.exit("Linear system is inconsistent.")
        
    
print "NEXT PHASE.\n\n\n"


while True:
    print "=== Phase 2", "Iteration", iteration, "\n"
#    print c
#    print basis_indices
#    print u
#    print A_b_0
    u = np.matrix([c[0, i] for i in basis_indices]) * A_b_0
    estimation = estimate2(input_data, u, A_b, c, basis_indices)
    print "Delta:", [(t[0], t[1]) for t in estimation]
    available_for_improvement = filter_estimates(input_data, estimation, alg.det(A_b))

    if len(available_for_improvement) == 0:
        print "No improvements are available.\n"
        print_result(c, values, input_data)
        sys.exit()

    
#    print "Working estimates: ", [(t[1], int(t[2])) for t in available_for_improvement]
    max_estimation_index = choose_estimation(available_for_improvement)
    print "Chosen:", "x" + str(max_estimation_index[1]), " with delta", max_estimation_index[2]

    AbAj = A_b_0 * extract_as_int(equations, max_estimation_index[1])
    teta_1, teta_2 = calculate_tetas(input_data, max_estimation_index, x_b, basis_indices, matrix_to_1D_array(AbAj), alg.det(A_b))
    print "Theta:", [("x" + str(i[1]), round(i[2], 5)) for i in teta_1], "and", [("x" + str(i[1]), round(i[2], 5)) for i in teta_2]

    if len(teta_1 + teta_2) == 0:
        print_result(c, values, input_data)
        sys.exit()
    elif all([i[2] == float('inf') for i in (teta_1 + teta_2)]):
        print "No non-infinitive theta was found."
        sys.exit("No solution have been found: objective function is not limited.")

    teta = choose_teta(teta_1, teta_2)
    print "Chosen theta:", teta[2], "for", "x" + str(teta[1])

    x_to_remove = basis_indices[teta[0]]
    x_to_add = max_estimation_index[1]
    print "Replacing", x_to_remove, " with ", x_to_add
    Q_base = np.concatenate((A_b, np.identity(A_b.shape[0], int), extract_as_int(equations, x_to_add)), axis=1	)
    Q0 = create_q_matrix(Q_base)
    Q1 = make_q_rotate(Q_base, Q0, teta[0])
    print "Building Q-matrix..."
    print Q_base
    print "..."
    print Q0
    print "Rotating column", teta[0], "..."
    print Q1
    print "Next basis matrix determinant =", Q1[teta[0],Q1.shape[1] - 1]

    pretty_print(x_b, basis_indices, A_b_0, u, AbAj, max_estimation_index[2], teta[2], max_estimation_index[1], (c * values[range(0, input_data.get_N()), 0])[0, 0] * alg.det(A_b))
    
    A_b[:,teta[0]:teta[0]+1] = extract_as_int(equations, x_to_add)
    A_b_ = Q1[:,[input_data.get_M() + i for i in range(0, input_data.get_M())]]
    if teta[4] == 1 and max_estimation_index[2] < 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_upper_value()
    elif teta[4] == 1 and max_estimation_index[2] > 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_lower_value()
    elif teta[4] == 2 and max_estimation_index[2] < 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_lower_value()
    elif teta[4] == 2 and max_estimation_index[2] > 0 :
        values[basis_indices[teta[0]]] = input_data.get_bound(teta[1]).get_upper_value()

    basis_indices[teta[0]] = max_estimation_index[1]
    
    non_basis_values = list()
    for i in range(0, values.shape[0]):
        if i not in basis_indices:
            if i >= input_data.get_N():
                non_basis_values.append(values[i, 0])
            else:
                non_basis_values.append(values[i, 0])
        else:
            non_basis_values.append(0)
    non_basis_values = np.matrix(non_basis_values).transpose()
    x_b_ = - int(np.sign(alg.det(A_b))) * A_b_ * (a * non_basis_values).astype(int)
    for i in range(0, x_b_.shape[0]):
        values[basis_indices[i]] = x_b_[i, 0] / abs(alg.det(A_b))
    x_b = x_b_
    A_b_0 = int(np.sign(alg.det(A_b))) * A_b_
    u[0, teta[0]] = 0
#    check(input_data, values)
    iteration = iteration + 1



